/* $Id: annealing.c,v 3.4 1998/06/01 16:10:04 kas Exp $ */

/* The Simulated Annealing (SA) algorithm */

/*
 *  This file is a part of Opal. See the Opal www page at
 *  http://www.fi.muni.cz/~kas/opal for details.
 *
 *  Copyright (C) 1995-1998  Jan "Yenya" Kasprzak <kas@fi.muni.cz>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <stdio.h>
#include <math.h>
#include <stdlib.h>	/* rand() family */
#include <signal.h>
#include <unistd.h>

#include "ff.h"
#include "pval.h"

#define INITIAL_TEMPERATURE	100.0
#ifndef NONTRANSFORMED_ONLY
#define MOVE_SCALE	3	/* Maximum difference of coordinates between
				 * two neighbour states.
				 */
#else
#define MOVE_SCALE	0.1
#endif

#define MOVE_SCALE_FACTOR	0.1
#define TIMEOUT		30	/* for alarm() call */
#define TEMP_CHANGE	0.9	/* temperature change factor */
#define MAX_ACCEPTED_FACTOR	15	/* average # of successful moves
				* with each atom sufficient to decrease
				* temperature.
				*/
#define MAX_FACTOR		1350	/* average # of all moves
				* with each atom sufficient to decrease
				* temperature.
				*/
#define FROZEN_STATE	5	/* After this # of MAX_FACTOR consequent
				* barriers reached the system is considered
				* frozen.
				*/
#ifndef NONTRANSFORMED_ONLY
#define LAMBDA0		0.8
#define LAMBDAN		0.011529
#define LAMBDA_FACTOR	0.8
#endif

#define MINVAL_EPSILON	0.25	/* What is a "reasonable" change of minimum */

/* #define DEBUG_ANNEALING	1 */

static void system_setup __P((potential_value *pvp));
static void temp_sig __P((int));
static void dump_sig __P((int));
static void save_minimum __P((potential_value *pvp));
static void dump_energy __P((potential_value *pvp));
static void change_temperature();
static void dump_parameters __P((potential_value *pvp));
#ifndef NONTRANSFORMED_ONLY
static void change_lambda __P((potential_value *pvp));
#endif

/* Implicit annealing parameters; changable via setup_annealing() */
static REAL initial_temperature = INITIAL_TEMPERATURE, move_scale = MOVE_SCALE,
	move_scale_factor = MOVE_SCALE_FACTOR, temp_factor = TEMP_CHANGE;
static int max_iterations_factor = MAX_FACTOR,
	max_accepted_factor = MAX_ACCEPTED_FACTOR;

static int flag_dump_energy = 0;
static int flag_change_temperature = 0;
static int flag_need_dump_outfile = 0;
static int done = 0;
static unsigned long max_iterations, max_accepted;
static int is_frozen = 0;
static unsigned long last_iterations = 0, last_accepted = 0;
static unsigned long iterations = 0, accepted = 0;
static char *outfile;
static REAL *x, *y, *z, minval;
static REAL temperature;
static REAL last_min;
static REAL current_move_scale;
#ifndef NONTRANSFORMED_ONLY
static char *ptmpl = NULL, *atmpl = NULL, *ttmpl = NULL;
static REAL lambda0 = LAMBDA0, lambdaN = LAMBDAN,
	lambda_factor = LAMBDA_FACTOR;
static REAL lambda = LAMBDA0;
#endif

void setup_annealing(of, t, ms, msf, maf, mf, tf
#ifndef NONTRANSFORMED_ONLY
	,l0, lN, lf, pt, at, tt
#endif
	)
char *of;
REAL t, ms, msf, tf;
int maf, mf;
#ifndef NONTRANSFORMED_ONLY
REAL l0, lN, lf;
char *pt, *at, *tt;
#endif
{
	if (of)		outfile = of;
	if (t)		initial_temperature = t;
	if (ms)		move_scale = ms;
	if (msf>=0.0)	move_scale_factor = msf;
	if (maf)	max_accepted_factor = maf;
	if (mf)		max_iterations_factor = mf;
	if (tf)		temp_factor = tf;
#ifndef NONTRANSFORMED_ONLY
	if (pt)		ptmpl = pt;
	if (at)		atmpl = at;
	if (tt)		ttmpl = tt;
	if (l0 >= 0.0)	lambda0 = l0;
	if (lN >= 0.0)	lambdaN = lN;
	if (lf >= 0.0)	lambda_factor = lf;
#endif
}

REAL annealing(pvp)
potential_value *pvp;
{

	system_setup(pvp);
	/* Compute the initial state and save this as first minimum */
	last_min = compute_potential(pvp);
	save_minimum(pvp);

	dump_parameters(pvp);

	printf ("Initial potential = %f\n", minval);


	while (!done)
	{
		int a;
		REAL x1, y1, z1, nval;

		iterations++;

#if 1
		if (!(iterations % 10000)) {
			char buf[1024];
#ifdef NONTRANSFORMED_ONLY
			fprintf(stderr, "%ld:%ld:%f:%f:", iterations, accepted,
				 pvp->value,minval);
#else
			fprintf(stderr, "%ld:%ld:%f:%f:%f:", iterations,
				accepted, lambda, pvp->value,minval);
#endif
			fflush(stderr);
			write_crd_file(pvp, outfile, x, y, z);
			sprintf(buf,"print_temp %s", outfile);
			system(buf);
		}
#endif
		/*
		 * Heuristics for changing the temperature:
		 * see Kirkpatrick83: p.675.
		 */
		if (accepted - last_accepted > max_accepted) {
			if (!is_frozen) {
				flag_change_temperature = 1;
				last_iterations = iterations;
			} else
				is_frozen = 0;
			last_accepted = accepted;
		}
		else if (iterations - last_iterations > max_iterations) {
			printf("Frozen = %d\n", ++is_frozen);
			if (is_frozen >= FROZEN_STATE) {
				printf("System considered frozen.\n");
				done = 1;
				continue;
			}
			flag_change_temperature = 1;
			last_iterations = iterations;
		}

		if (flag_dump_energy)
			dump_energy(pvp);

		if (flag_change_temperature) {
			change_temperature();
#ifndef NONTRANSFORMED_ONLY
			change_lambda(pvp);
#endif
		}
			
		/* Compute next state */
		a = rand() % pvp->ffp->natoms;
		x1 = (double)(rand()-RAND_MAX/2)*current_move_scale/(RAND_MAX/2);
		y1 = (double)(rand()-RAND_MAX/2)*current_move_scale/(RAND_MAX/2);
		z1 = (double)(rand()-RAND_MAX/2)*current_move_scale/(RAND_MAX/2);
#ifdef DEBUG_ANNEALING
		printf ("Annealing: Trying to move atom %d to [%f,%f,%f]\n",
			a+1, x1, y1, z1);
#endif
		x1 += pvp->x[a];
		y1 += pvp->y[a];
		z1 += pvp->z[a];

		nval = recompute_potential(pvp, a, x1, y1, z1);

		if (nval > pvp->ovalue)
		{
#ifdef DEBUG_ANNEALING
			REAL prob, rnd;
			prob = exp((pvp->ovalue-nval)/temperature);
			rnd = (double)rand()/RAND_MAX;
			printf("DeltaE = %f, Probability = %f, Rand = %f\n",
				nval - pvp->ovalue, prob, rnd);
			if (prob < rnd)
#else
			if (exp((pvp->ovalue-nval)/temperature)
				< (double)rand()/RAND_MAX)
#endif
			{
#ifdef DEBUG_ANNEALING
				printf ("Rejecting state w/ energy %f (old=%f)\n",
					nval, pvp->ovalue);
#endif
				undo_new_potential(pvp);
				continue;
			}
		}
		/* We have accepted the new state; Now we have to look
		 * whether minimum is improved by this new state.
		 */
		accepted++;
#ifdef DEBUG_ANNEALING
		printf("Accepted the new state.\n");
#endif
		if (nval < minval)
		{
#ifdef DEBUG_ANNEALING
			printf("New minimum: %3.3f\n", nval);
#endif
			/* System is no longer considered frozen,
			 * if there was a reasonable change of minval.
			 */
			if (last_min - nval > MINVAL_EPSILON) {
				is_frozen = 0;
				last_iterations = iterations;
				last_min = nval;
			}

			save_minimum(pvp);

			flag_need_dump_outfile = 1;
#ifdef DEBUG_ANNEALING
			printf("Veryfying: %f\n", compute_potential(pvp));
#endif
		}
	}

	free(pvp->x);
	free(pvp->y);
	free(pvp->z);
	pvp->x = x;
	pvp->y = y;
	pvp->z = z;
	compute_potential(pvp);
	dump_energy(pvp);
#if 1
	fprintf(stderr,"%f:%f:%f:%d:%d:%f:%f:%ld:%ld:%f:",
		initial_temperature, move_scale, move_scale_factor,
		max_accepted_factor, max_iterations_factor, temp_factor,
		temperature, iterations, accepted, pvp->value);
	fflush(stderr);
	{
		char buf[1024];
		sprintf(buf,"print_temp %s", outfile);
		system(buf);
	}
#endif
	return pvp->value;
}


static void system_setup(pvp)
potential_value *pvp;
{
	/* Allocate space for current optimum */
	x = malloc(sizeof(REAL)*pvp->ffp->natoms);
	y = malloc(sizeof(REAL)*pvp->ffp->natoms);
	z = malloc(sizeof(REAL)*pvp->ffp->natoms);

	/* Compute parameters for the procedure of the temperature changing */
	max_iterations = max_iterations_factor*pvp->ffp->natoms;
	max_accepted = max_accepted_factor*pvp->ffp->natoms;
	temperature = initial_temperature;
	current_move_scale = move_scale;
#ifndef NONTRANSFORMED_ONLY
	lambda = lambda0;
	read_trans_ff(pvp->ffp, lambda, atmpl, ttmpl, ptmpl);
#endif
	signal(SIGTERM, dump_sig);
	signal(SIGHUP, dump_sig);
	signal(SIGUSR1, temp_sig);
	signal(SIGUSR2, temp_sig);
	signal(SIGALRM, dump_sig);
	alarm(TIMEOUT);
}

static void dump_sig(sig)
int sig;
{
	signal(sig, dump_sig);
	flag_dump_energy = sig;
	if (sig == SIGHUP)
		flag_need_dump_outfile = 1;
	if (sig == SIGALRM)
		alarm(TIMEOUT);
}

static void temp_sig(sig)
int sig;
{
	signal(sig, temp_sig);
	flag_change_temperature = sig == SIGUSR1 ? 1 : -1;
}

static void save_minimum(pvp)
potential_value *pvp;
{
	minval = pvp->value;
	memcpy(x, pvp->x, sizeof(REAL)*pvp->ffp->natoms);
	memcpy(y, pvp->y, sizeof(REAL)*pvp->ffp->natoms);
	memcpy(z, pvp->z, sizeof(REAL)*pvp->ffp->natoms);
}

void dump_energy(pvp)
potential_value *pvp;
{
	printf("Energy: current = %3.3f, min = %3.3f\n\tafter %lu iterations, %lu accepted\n",
		pvp->value, minval, iterations, accepted);
	printf("\tat temperature %f, move_scale %f\n",
		temperature, current_move_scale);

	if (flag_dump_energy == SIGTERM)
	{
		printf("Terminating on SIGTERM\n");
		done = 1;
	}
	flag_dump_energy = 0;
	if (flag_need_dump_outfile && outfile)
		write_crd_file(pvp, outfile, x, y, z);
	flag_need_dump_outfile = 0;
}

static void change_temperature()
{
	if (flag_change_temperature == 1) {
		temperature *= temp_factor;
		current_move_scale -= move_scale_factor;
	} else {
		temperature /= temp_factor;
		current_move_scale += move_scale_factor;
	}
	flag_change_temperature = 0;
	if (current_move_scale < 0.1)
		current_move_scale = 0.1;
#ifdef DEBUG_ANNEALING
	printf("Temperature changed to %3.3f, move_scale to %f\n",
		temperature, current_move_scale);
#endif
}

static void dump_parameters(pvp)
potential_value *pvp;
{
	printf("temp=%f; move_scale=%f; m_s_factor=%f, max_acc_f=%d; max_i_f=%d; temp_chg=%f\n",
		initial_temperature, move_scale, move_scale_factor,
		max_accepted_factor, max_iterations_factor, temp_factor);
	printf("Annealing:\n%s\n", pvp->ffp->title);
}

#ifndef NONTRANSFORMED_ONLY
static void change_lambda(pvp)
potential_value *pvp;
{
	static int reached_msg = 0;
	static int last_change = 20;
	REAL oldE = minval;

	if (--last_change>0)
		return;

	lambda *= lambda_factor;

	if (lambda <= lambdaN) {
		lambda /= lambda_factor;
#ifdef CONTINUE_WITH_LAST_LAMBDA
		if (!reached_msg) {
			printf("Reached lambdaN.\n");
			reached_msg = 1;
		}
#else
		printf("Reached lambdaN (lambda=%f).\n", lambda);
		done = 1;
#endif
		return;
	}
	last_change = 2;

	memcpy(pvp->x, x, sizeof(REAL)*pvp->ffp->natoms);
	memcpy(pvp->y, y, sizeof(REAL)*pvp->ffp->natoms);
	memcpy(pvp->z, z, sizeof(REAL)*pvp->ffp->natoms);

	printf("Running local minimization:\n");
	printf("Local minimum: %f\n", minimize_l(pvp,LMIN_ACCURACY));

	memcpy(x, pvp->x, sizeof(REAL)*pvp->ffp->natoms);
	memcpy(y, pvp->y, sizeof(REAL)*pvp->ffp->natoms);
	memcpy(z, pvp->z, sizeof(REAL)*pvp->ffp->natoms);

	write_crd_file(pvp, "lambda.crd", x, y, z);
	oldE = compute_potential(pvp);
	read_trans_ff(pvp->ffp, lambda, atmpl, ttmpl, ptmpl);
	last_min = compute_potential(pvp);
	save_minimum(pvp);
	printf("Lambda changed to %f, energy=%f (was %f)\n",
		lambda, pvp->value, oldE);
}
#endif
