/* $Id: lmin_main.c,v 3.10 1998/06/01 16:10:12 kas Exp $ */

/* Main function for local-minimizing program */

/*
 *  This file is a part of Opal. See the Opal www page at
 *  http://www.fi.muni.cz/~kas/opal for details.
 *
 *  Copyright (C) 1995-1998  Jan "Yenya" Kasprzak <kas@fi.muni.cz>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include "config.h"
#ifdef HAVE_GETOPT_LONG
#include <getopt.h>
#else
#include "getopt.h"
#endif

#include "ff.h"
#include "pval.h"

static char *program_name;

static void usage();
#ifndef NONTRANSFORMED_ONLY
static void minimize_l_t __P((potential_value *pvp, REAL accuracy,
	REAL lambda0, REAL lambdaN, REAL lambda_factor,
	char *atmpl, char *ttmpl, char *ptmpl));
#endif

int main(argc,argv)
int argc;
char **argv;
{
	REAL vdw14 = -1.0, eel14 = -1.0, diel_lin = -1.0, diel_const = -1.0;
	REAL cutoff = -1.0;
	REAL accuracy = 1.0e-5;
	int numeric = 0;
	force_field *ffp;
	potential_value *pvp;
	char *outfile = NULL;
	char *infile = NULL;
#ifndef NONTRANSFORMED_ONLY
	char *ptmpl = NULL, *atmpl = NULL, *ttmpl = NULL;
	REAL lambda0=-1, lambdaN=-1, lambda_factor=-1;
#endif
	program_name = *argv;
	lmin_verbose = 2;

	while(1)
	{
		int c, opt_ind;
		static struct option long_opts[] = {
			{ "help", 0, 0, 'h' },
			{ "version", 0, 0, 'V' },
			{ "vdw14", 1, 0, 'w' },
			{ "eel14", 1, 0, 'e' },
			{ "diel-lin", 1, 0, 'l' },
			{ "diel-const", 1, 0, 'c' },
			{ "cutoff", 1, 0, 'C' },
			{ "seed", 1, 0, 's' },
			{ "outfile", 1, 0, 'o' },
			{ "begin-coords", 1, 0, 'b' },
			{ "numgrad", 0, 0, 'n' },
			{ "verbose", 1, 0, 'v' },
			{ "accuracy", 1, 0, 'a' },
#ifndef NONTRANSFORMED_ONLY
			{ "pair-template", 1, 0, 'P' },
			{ "angle-template", 1, 0, 'A' },
			{ "dih-template", 1, 0, 'D' },
#endif
			{ NULL }
		};

		if ((c=getopt_long(argc, argv, "hvw:a:e:l:c:s:o:b:--", long_opts,
			&opt_ind)) == -1)
			break;
		switch(c)
		{
		case 'h':
		case '?':
			usage();
		case 'V':
			printf("Version: $Id: lmin_main.c,v 3.10 1998/06/01 16:10:12 kas Exp $\n");
			return 0;
		case 'a':
			accuracy = atof(optarg);
			break;
		case 'w':
			vdw14 = atof(optarg);
			break;
		case 'e':
			eel14 = atof(optarg);
			break;
		case 'l':
			diel_lin = atof(optarg);
			break;
		case 'c':
			diel_const = atof(optarg);
			break;
		case 'C':
			cutoff = atof(optarg);
			break;
		case 's':
			srand(atoi(optarg));
			break;
		case 'o':
			outfile = optarg;
			break;
		case 'b':
			infile = optarg;
			break;
		case 'n':
			numeric = 1;
			break;
		case 'v':
			lmin_verbose = atoi(optarg);
			break;
#ifndef NONTRANSFORMED_ONLY
		case 'P':
			ptmpl = optarg;
			break;
		case 'A':
			atmpl = optarg;
			break;
		case 'D':
			ttmpl = optarg;
			break;
                case 'L':
                        lambda0 = atof(optarg);
                        break;
                case 'E':
                        lambdaN = atof(optarg);
                        break;
                case 'F':
                        lambda_factor = atof(optarg);
                        break;
#endif
		default:
			usage();
			return -1;
		}
	}

	if (optind+1 > argc)
		usage();

	if ((ffp = prm2ff(argv[optind])) == NULL)
	{
		fprintf(stderr, "Cannot read force field from %s\n",
			argv[optind]);
		return 1;
	}

	setup_ff(ffp, vdw14, eel14, diel_lin, diel_const, cutoff);
	pvp = alloc_pv(ffp);

        if (infile)
                read_crd_file(pvp, infile);
        else
		generate_coordinates(pvp);

#ifndef NONTRANSFORMED_ONLY
	minimize_l_t(pvp, accuracy, lambda0, lambdaN, lambda_factor, atmpl,
		ttmpl, ptmpl);
#else
	minimize_l(pvp, accuracy, numeric); 
#endif
	if (outfile)
		write_crd_file(pvp, outfile, pvp->x, pvp->y, pvp->z);
	if (lmin_verbose >= 2)
		printf("Minimum is ");
	if (lmin_verbose >= 1)
		printf("%f\n", pvp->value);

	return 0;
}

static void usage()
{
	fprintf(stderr, "Usage: %s [options] file.top\n"
		"Where [options] is zero or more from\n"
		"--help\tprints out this message\n"
		"--version\tprints out version of program\n"
		"--vdw14 number\tsets the scaling parameter for VdW 1-4 interactions\n"
		"--eel14 number\tsets the scaling parameter for E-static 1-4 interactions\n"
		"--diel-lin number\tsets the linear part of dielectric funtion\n"
		"--diel-const number\tsets the constant part of dielectric function\n"
		"--cutoff number\tsets the 1991 force field cutoff\n"
		"--seed number\tseed for the random number generator\n"
		"--outfile filename.crd\twhere coordinates of minimum state will be written\n"
		"--begin-coords filename.crd\tstarting state of the molecule\n"
		"--accuracy number\taccuracy for the local minimizing procedure\n"
		"--numgrad\tuse gradient computed by finite differeces (slower!)\n"
#ifndef NONTRANSFORMED_ONLY
		"--pair-template string\ttemplate for pair-interactions file\n"
		"--angle-template string\ttemplate for angle-interactions file\n"
		"--dih-template string\ttemplate for torsion-interactions file\n"
#endif
		,program_name);
	exit(1);
}

#ifndef NONTRANSFORMED_ONLY
static void minimize_l_t(pvp, accuracy, lambda0, lambdaN, lambda_factor,
	atmpl, ttmpl, ptmpl)
potential_value *pvp;
REAL accuracy, lambda0, lambdaN, lambda_factor;
char *atmpl, *ttmpl, *ptmpl;
{
	REAL lambda = lambda0;

	do {
		if (!read_trans_ff(pvp->ffp, lambda, atmpl, ttmpl, ptmpl)) {
			fprintf(stderr, "Cannot read file\n");
			exit (1);
		}
		minimize_l(pvp, accuracy, 1);
		lambda /= lambda_factor;
		free_trans_part(pvp->ffp);
	} while (lambda > lambdaN);
}
#endif
