/* $Id: nontrans.c,v 3.17 1998/06/01 16:10:12 kas Exp $ */

/* Functions for computing components of the non-transformed force field */

/*
 *  This file is a part of Opal. See the Opal www page at
 *  http://www.fi.muni.cz/~kas/opal for details.
 *
 *  Copyright (C) 1995-1998  Jan "Yenya" Kasprzak <kas@fi.muni.cz>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <stdio.h>
#include <math.h>

#include "ff.h"
#include "linalg.h"
#ifdef PVAL_VERBOSE
/* Definition of pval_verbose variable */
#include "pval.h"
#endif

/*
 * Hopefuly the `r' variable below will be optimized out unless PVAL_VERBOSE
 * is defined.
 */
REAL compute_bond(ffp, x, y, z, bond)
force_field *ffp;
REAL *x, *y, *z;
int bond;
{
	REAL r, b;
	int at1, at2;

	at1 = ffp->bondat1[bond];
	at2 = ffp->bondat2[bond];

	r =  sqrt((x[at1] - x[at2])*(x[at1] - x[at2])
		+ (y[at1] - y[at2])*(y[at1] - y[at2])
		+ (z[at1] - z[at2])*(z[at1] - z[at2]));
	b = r - ffp->req[ffp->bondnum[bond]];
	b = ffp->rk[ffp->bondnum[bond]] * b*b;
#ifdef PVAL_VERBOSE
	if (pval_verbose >= PVAL_DETAILED)
		printf("BOND %4d-%4d %4d %7.3f %7.3f %7.3f          %7.3f\n",
			at1+1, at2+1, ffp->bondnum[bond],
			ffp->rk[ffp->bondnum[bond]],
			ffp->req[ffp->bondnum[bond]], r, b);
#endif
	return b;
}

REAL compute_angle(ffp, x, y, z, angle)
force_field *ffp;
REAL *x, *y, *z;
int angle;
{
	REAL a, e; /* e will be optimized out unless PVAL_VERBOSE is defined */
	int at1, at2, at3;
	REAL x12[3], x32[3];

	at1 = ffp->angleat1[angle];
	at2 = ffp->angleat2[angle];
	at3 = ffp->angleat3[angle];

	if (x[at1] - x[at2] == 0.0 && y[at1] - y[at2] == 0.0
		&& z[at1] - z[at2] == 0.0)
	{
		a = -1.0;
		goto computed;
	}

	if (x[at3] - x[at2] == 0.0 && y[at3] - y[at2] == 0.0
		&& z[at3] - z[at2] == 0.0)
	{
		a = -1.0;
		goto computed;
	}

		
	vectpts(x12,x,y,z,at1,at2);
	vectpts(x32,x,y,z,at3,at2);

	a = dot3(x12,x32) / norm3(x12) / norm3(x32);

	/*
	 * FIXME: Tady se mozna bude muset zohlednit, jestli ten
	 * rozdil nepreleze ven z intervalu [-pi,pi].
	 */

	a = acos(a);
computed:
	e = a - ffp->teq[ffp->anglenum[angle]];
	e = ffp->tk[ffp->anglenum[angle]]*e*e;

#ifdef PVAL_VERBOSE
	if (pval_verbose >= PVAL_DETAILED)
		printf ("ANGL %4d-%4d-%4d %4d %7.3f %7.3f %7.3f     %7.3f\n",
			at1+1, at2+1, at3+1, ffp->anglenum[angle],
			ffp->tk[ffp->anglenum[angle]],
			(double)(ffp->teq[ffp->anglenum[angle]]*180.0/M_PI),
			(double)(a*180.0/M_PI), e);
#endif
	return e;
}

REAL compute_dih(ffp, x, y, z, dih)
force_field *ffp;
REAL *x, *y, *z;
int dih;
{
	int a1, a2, a3, a4;
	REAL a[3], b[3], c[3], d[3];
	REAL total, cos_phi, n1, n2;
	int n, dihnum;

	a1 = ffp->dihat1[dih];
	a2 = ffp->dihat2[dih];
	a3 = ffp->dihat3[dih];
	a4 = ffp->dihat4[dih];
	dihnum = ffp->dihnum[dih];
	/* Normalovy vektor k rovine 1-2-3 */
	vectpts(a,x,y,z,a1,a2);
	vectpts(b,x,y,z,a3,a2);
	cross3(c,a,b);
	n1 = norm3(c);

	/* Normalovy vektor k rovine 2-3-4 */
	vectpts(a,x,y,z,a4,a3);
	cross3(d,a,b);
	n2 = norm3(d);

	if (!(n1*=n2))
		return -1.0;

	cos_phi = dot3(c,d) / n1;

	/*
	 * Now in cos_phi we have a value of cos(<(1-2-3,2-3-4)).
	 * The following code makes cos(Pn*<(1-2-3,2-3-4) - Phase) from it.
	 */

	total = 0.0;
	do {
		REAL temp;
		if ((n=ffp->pn[dihnum]) < 0) {
			n = -n;
		}
		/* compute temp = cos(n*phi) from cos_phi */
		switch (n)
		{
		case 1:
			temp = cos_phi;
			break;
		case 2:	/* cos(2x) = 2cos^2(x)-1 */
			temp = 2*cos_phi*cos_phi-1.0;
			break;
		case 3:	/* cos(3x) = 4cos^3(x)-3cos(x) */
			temp = 4*cos_phi*cos_phi*cos_phi-3*cos_phi;
			break;
		case 4: /* cos(4x) = 8cos^4(x)-8cos^2(x)+1 */
			temp = 8*cos_phi*cos_phi*cos_phi*cos_phi
				- 8*cos_phi*cos_phi + 1;
			break;
		default:
			fprintf(stderr, "FATAL: in compute_dih(): Illegal periodicity (%d).\n",n);
			exit(1);
		}

		total += temp = ffp->pk[dihnum]*(1 + ffp->phase[dihnum]*temp);
#ifdef PVAL_VERBOSE
	if (pval_verbose >= PVAL_DETAILED)
		printf ("%4s %4d-%4d-%4d-%4d %4d %5.0f %7.3f %7.3f  %7.3f\n",
			ffp->improper[dihnum]?"ITOR":"TORS",
			a1+1, a2+1, a3+1, a4+1,
			dihnum, ffp->phase[dihnum] == 1 ? 0.0 : 180.0,
			ffp->pk[dihnum], (double)(acos(cos_phi)*180/M_PI), temp);
#endif
	} while (ffp->pn[dihnum++] < 0);

	if (!ffp->ignored[dih])
		total += compute_14(ffp, x,y,z, a1,a4);
	return total;
}

REAL compute_nb(ffp, x,y,z, a1,a2)
force_field *ffp;
REAL *x, *y, *z;
int a1, a2;
{
	REAL r, vdw, eel;
	int idx;
#ifdef PVAL_VERBOSE
	int hbond;
#endif
	idx = ffp->nbnum[ffp->natypes*ffp->atypes[a1]+ffp->atypes[a2]];
#ifdef PVAL_VERBOSE
	hbond=(idx<0?1:0);
#endif
	r = (x[a1]-x[a2])*(x[a1]-x[a2])
		+(y[a1]-y[a2])*(y[a1]-y[a2])
		+(z[a1]-z[a2])*(z[a1]-z[a2]);

	/* Momentln je v `r' hodnota r^2. Odmocnme pozdji. */

	/* FIXME: Tady by mlo sp bt r<epsilon pro njak epsilon */
	if (r == 0.0) {
		vdw = eel = 1E+20;
		/*
		 * Nsleduje skok, ale pedpokldm, e pokud nebude
		 * definovno PVAL_VERBOSE, tak to kompiltor zoptimalizuje
		 * bez skoku.
		 */
		goto print_and_return;
	}

	/*
	 * Otestujeme, jestli jde o van der Waalsovskou interakci
	 * (6-12) nebo o vodikovou vazbu (10-12). Pokud je idx < 0,
	 * jde o vodikovou vazbu. Pak je nutne snizit index o dva:
	 * Puvodne bylo pole (Cno ve struct parm) indexovano tak,
	 * ze 1,2,... znamenalo vdw, a -1,-2,... znamenalo vodikovou
	 * vazbu. Protoze vsechny indexove pole snizuji o jednicku
	 * (aby to sedelo na C-style pole), je momentalne v idx
	 * hodnota s vyznamem: 0,1,2,... -- vdw, -2,-3... -- vodikova-2.
	 * FIXME: Existuje nejake elegantnejsi reseni tohoto problemu?
	 */
	if (idx >= 0)
		vdw = ffp->r12[idx]/(r*r*r*r*r*r)
			- ffp->r6[idx]/(r*r*r);
	else
		vdw = ffp->h12[-idx-2]/(r*r*r*r*r*r)
			- ffp->h10[-idx-2]/(r*r*r*r*r);

	r = sqrt(r);
	eel = ffp->charge[a1]*ffp->charge[a2]
		/ ((ffp->diel_lin*r + ffp->diel_const)*r);
print_and_return:
#ifdef PVAL_VERBOSE
	if (pval_verbose >= PVAL_DETAILED)
		printf("%4s %4d-%4d %12.3f %8.3f %7.3f %7.3f %7.3f\n",
			hbond?"HBND":"NBND",
			a1+1,a2+1,
			(hbond?ffp->h12[-idx-2]:ffp->r12[idx]),
			(hbond?ffp->h10[-idx-2]:ffp->r6[idx]), r,
			vdw, eel);
#endif
	return vdw + eel;
}

REAL compute_14(ffp, x,y,z, a1,a2)
force_field *ffp;
REAL *x, *y, *z;
int a1, a2;
{
	REAL r, vdw, eel;
	int idx;

	idx = ffp->nbnum[ffp->natypes*ffp->atypes[a1]+ffp->atypes[a2]];
	/*
	 * Vypocet pro 1-4 interakce je trochu odlisny.
	 * Pocita se zde VdW interakce (12-6) i pro vodikove
	 * atomy. Index se v tomto pripade ziska jako
	 * vetsi_cislo*(vetsi_cislo+1)/2+mensi_cislo.
	 */
	if (idx < 0) {
		int tsmall, tbig, foo;
		tsmall=ffp->atypes[a1];
		tbig=ffp->atypes[a2];
		if (tsmall > tbig) {
			foo = tsmall; tsmall = tbig; tbig = foo;
		}
		idx = (tbig*(tbig+1))/2 + tsmall;
	}

	r = compute_distance_sq(x,y,z, a1, a2);
	/* Momentln je v `r' hodnota r^2. Odmocnme pozdji. */

	if (r == 0.0) {
		vdw = eel = 1E+20;
		goto print_return;
	}

	vdw = ffp->r12[idx]/(r*r*r*r*r*r)
		- ffp->r6[idx]/(r*r*r);
	vdw /= ffp->vdw_14;

	r = sqrt(r);
	eel = ffp->charge[a1]*ffp->charge[a2]
		/ ((ffp->diel_lin*r + ffp->diel_const)*r);
	eel /= ffp->eel_14;

print_return:
#ifdef PVAL_VERBOSE
	if (pval_verbose >= PVAL_DETAILED)
		printf("D1-4 %4d-%4d %12.3f %7.3f %7.3f %7.3f %8.3f\n",
			a1+1,a2+1,
			ffp->r12[idx],
			ffp->r6[idx], r,
			vdw, eel);
#endif
	return vdw + eel;
}

